from __future__ import division
from __future__ import unicode_literals

import json
from builtins import str
from copy import deepcopy

from future.utils import itervalues, iteritems
from snips_nlu_ontology import get_all_languages

from snips_nlu.builtin_entities import is_builtin_entity
from snips_nlu.constants import (TEXT, USE_SYNONYMS, SYNONYMS, DATA, INTENTS,
                                 ENTITIES, ENTITY, SLOT_NAME, UTTERANCES,
                                 LANGUAGE, VALUE, AUTOMATICALLY_EXTENSIBLE,
                                 CAPITALIZE, VALIDATED)
from snips_nlu.string_variations import get_string_variations
from snips_nlu.tokenization import tokenize_light
from snips_nlu.utils import validate_type, validate_key, validate_keys


def extract_queries_entities(dataset):
    entities_values = {ent_name: [] for ent_name in dataset[ENTITIES]}

    for intent in itervalues(dataset[INTENTS]):
        for query in intent[UTTERANCES]:
            for chunk in query[DATA]:
                if ENTITY in chunk and not is_builtin_entity(chunk[ENTITY]):
                    entities_values[chunk[ENTITY]].append(chunk[TEXT])
    return {k: list(v) for k, v in iteritems(entities_values)}


def validate_and_format_dataset(dataset):
    """Checks that the dataset is valid and format it"""
    # Make this function idempotent
    if dataset.get(VALIDATED, False):
        return dataset
    dataset = deepcopy(dataset)
    dataset = json.loads(json.dumps(dataset))
    validate_type(dataset, dict)
    mandatory_keys = [INTENTS, ENTITIES, LANGUAGE]
    for key in mandatory_keys:
        validate_key(dataset, key, object_label="dataset")
    validate_type(dataset[ENTITIES], dict)
    validate_type(dataset[INTENTS], dict)
    language = dataset[LANGUAGE]
    validate_type(language, str)
    if language not in get_all_languages():
        raise ValueError("Unknown language: '%s'" % language)

    for intent in itervalues(dataset[INTENTS]):
        validate_and_format_intent(intent, dataset[ENTITIES])

    queries_entities_values = extract_queries_entities(dataset)

    for entity_name, entity in iteritems(dataset[ENTITIES]):
        if is_builtin_entity(entity_name):
            dataset[ENTITIES][entity_name] = \
                validate_and_format_builtin_entity(entity)
        else:
            dataset[ENTITIES][entity_name] = validate_and_format_custom_entity(
                entity, queries_entities_values[entity_name], language)
    dataset[VALIDATED] = True
    return dataset


def validate_and_format_intent(intent, entities):
    validate_type(intent, dict)
    validate_key(intent, UTTERANCES, object_label="intent dict")
    validate_type(intent[UTTERANCES], list)
    for utterance in intent[UTTERANCES]:
        validate_type(utterance, dict)
        validate_key(utterance, DATA, object_label="utterance")
        validate_type(utterance[DATA], list)
        for chunk in utterance[DATA]:
            validate_type(chunk, dict)
            validate_key(chunk, TEXT, object_label="chunk")
            if ENTITY in chunk or SLOT_NAME in chunk:
                mandatory_keys = [ENTITY, SLOT_NAME]
                validate_keys(chunk, mandatory_keys, object_label="chunk")
                if is_builtin_entity(chunk[ENTITY]):
                    continue
                else:
                    validate_key(entities, chunk[ENTITY],
                                 object_label=ENTITIES)
    return intent


def get_text_from_chunks(chunks):
    return ''.join(chunk[TEXT] for chunk in chunks)


def has_any_capitalization(entity_utterances, language):
    for utterance in entity_utterances:
        tokens = tokenize_light(utterance, language)
        if any(t.isupper() or t.istitle() for t in tokens):
            return True
    return False


def add_variation_if_needed(utterances, variation, utterance, language):
    if not variation:
        return utterances
    all_variations = get_string_variations(variation, language)
    for v in all_variations:
        if v not in utterances:
            utterances[v] = utterance
    return utterances


def validate_and_format_custom_entity(entity, queries_entities, language):
    validate_type(entity, dict)
    mandatory_keys = [USE_SYNONYMS, AUTOMATICALLY_EXTENSIBLE, DATA]
    validate_keys(entity, mandatory_keys, object_label="entity")
    validate_type(entity[USE_SYNONYMS], bool)
    validate_type(entity[AUTOMATICALLY_EXTENSIBLE], bool)
    validate_type(entity[DATA], list)

    formatted_entity = dict()
    formatted_entity[AUTOMATICALLY_EXTENSIBLE] = entity[
        AUTOMATICALLY_EXTENSIBLE]
    use_synonyms = entity[USE_SYNONYMS]

    # Validate format and filter out unused data
    valid_entity_data = []
    for entry in entity[DATA]:
        validate_type(entry, dict)
        validate_keys(entry, [VALUE, SYNONYMS], object_label="entity entry")
        entry[VALUE] = entry[VALUE].strip()
        if not entry[VALUE]:
            continue
        validate_type(entry[SYNONYMS], list)
        entry[SYNONYMS] = [s.strip() for s in entry[SYNONYMS]
                           if len(s.strip()) > 0]
        valid_entity_data.append(entry)
    entity[DATA] = valid_entity_data

    # Compute capitalization before normalizing
    # Normalization lowercase and hence lead to bad capitalization calculation
    formatted_entity[CAPITALIZE] = has_any_capitalization(queries_entities,
                                                          language)

    # Normalize
    validated_data = dict()
    for entry in entity[DATA]:
        entry_value = entry[VALUE]
        validated_data = add_variation_if_needed(
            validated_data, entry_value, entry_value, language)

        if use_synonyms:
            for s in entry[SYNONYMS]:
                validated_data = add_variation_if_needed(
                    validated_data, s, entry_value, language)

    formatted_entity[UTTERANCES] = validated_data
    # Merge queries_entities
    for value in queries_entities:
        formatted_entity = add_entity_value_if_missing(
            value, formatted_entity, language)

    return formatted_entity


def validate_and_format_builtin_entity(entity):
    validate_type(entity, dict)
    return entity


def add_entity_value_if_missing(value, entity, language):
    entity[UTTERANCES] = add_variation_if_needed(entity[UTTERANCES], value,
                                                 value, language)
    return entity



from .config import Config, ProcessingUnitConfig
from .features import default_features_factories
from .intent_classifier import (
    LogRegIntentClassifierConfig, IntentClassifierDataAugmentationConfig,
    FeaturizerConfig)
from .intent_parser import (DeterministicIntentParserConfig,
                            ProbabilisticIntentParserConfig)
from .nlu_engine import NLUEngineConfig
from .slot_filler import CRFSlotFillerConfig, SlotFillerDataAugmentationConfig



from __future__ import unicode_literals

import errno
import numbers
import os
from builtins import object
from collections import OrderedDict, namedtuple, Mapping

import numpy as np

from snips_nlu.constants import (INTENTS, UTTERANCES, DATA, SLOT_NAME, ENTITY,
                                 RESOURCES_PATH, END, START)

REGEX_PUNCT = {'\\', '.', '+', '*', '?', '(', ')', '|', '[', ']', '{', '}',
               '^', '$', '#', '&', '-', '~'}


class NotTrained(LookupError):
    """Exception used when a processing unit is used while not fitted"""
    pass


class ClassPropertyDescriptor(object):
    def __init__(self, fget, fset=None):
        self.fget = fget
        self.fset = fset

    def __get__(self, obj, klass=None):
        if klass is None:
            klass = type(obj)
        return self.fget.__get__(obj, klass)()

    def __set__(self, obj, value):
        if not self.fset:
            raise AttributeError("can't set attribute")
        type_ = type(obj)
        return self.fset.__get__(obj, type_)(value)

    def setter(self, func):
        if not isinstance(func, (classmethod, staticmethod)):
            func = classmethod(func)
        self.fset = func
        return self


def classproperty(func):
    if not isinstance(func, (classmethod, staticmethod)):
        func = classmethod(func)

    return ClassPropertyDescriptor(func)


# pylint: enable=C0103


def type_error(expected_type, found_type):
    return TypeError("Expected %s but found: %s" % (expected_type, found_type))


def validate_type(obj, expected_type):
    if not isinstance(obj, expected_type):
        raise type_error(expected_type=expected_type, found_type=type(obj))


def missing_key_error(key, object_label=None):
    if object_label is None:
        return KeyError("Missing key: '%s'" % key)
    return KeyError("Expected %s to have key: '%s'" % (object_label, key))


def validate_key(obj, key, object_label=None):
    if key not in obj:
        raise missing_key_error(key, object_label)


def validate_keys(obj, keys, object_label=None):
    for key in keys:
        validate_key(obj, key, object_label)


def validate_range(rng):
    if not isinstance(rng, (list, tuple)) or len(rng) != 2 or rng[0] > rng[1]:
        raise ValueError("range must be a length 2 list or tuple and must be "
                         "valid")


class LimitedSizeDict(OrderedDict):
    def __init__(self, *args, **kwds):
        if "size_limit" not in kwds:
            raise ValueError("'size_limit' must be passed as a keyword "
                             "argument")
        self.size_limit = kwds.pop("size_limit")
        if len(args) > 1:
            raise TypeError('expected at most 1 arguments, got %d' % len(args))
        if len(args) == 1 and len(args[0]) + len(kwds) > self.size_limit:
            raise ValueError("Tried to initialize LimitedSizedDict with more "
                             "value than permitted with 'limit_size'")
        super(LimitedSizeDict, self).__init__(*args, **kwds)

    def __setitem__(self, key, value, dict_setitem=OrderedDict.__setitem__):
        dict_setitem(self, key, value)
        self._check_size_limit()

    def _check_size_limit(self):
        if self.size_limit is not None:
            while len(self) > self.size_limit:
                self.popitem(last=False)

    def __eq__(self, other):
        if self.size_limit != other.size_limit:
            return False
        return super(LimitedSizeDict, self).__eq__(other)


class UnupdatableDict(dict):
    def __setitem__(self, key, value):
        if key in self:
            raise KeyError("Can't update key '%s'" % key)
        super(UnupdatableDict, self).__setitem__(key, value)


def namedtuple_with_defaults(typename, field_names, default_values=()):
    T = namedtuple(typename, field_names)  # pylint: disable=C0103
    T.__new__.__defaults__ = (None,) * len(T._fields)
    if isinstance(default_values, Mapping):
        prototype = T(**default_values)
    else:
        prototype = T(*default_values)
    T.__new__.__defaults__ = tuple(prototype)
    return T


def get_resources_path(language):
    return os.path.join(RESOURCES_PATH, language)


def mkdir_p(path):
    """Reproduces the 'mkdir -p shell' command

    See
    http://stackoverflow.com/questions/600268/mkdir-p-functionality-in-python
    """
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


# pylint:disable=line-too-long
def regex_escape(s):
    """Escapes all regular expression meta characters in *s*

    The string returned may be safely used as a literal in a regular
    expression.

    This function is more precise than :func:`re.escape`, the latter escapes
    all non-alphanumeric characters which can cause cross-platform
    compatibility issues.

    References:

    - https://github.com/rust-lang/regex/blob/master/regex-syntax/src/lib.rs#L1685
    - https://github.com/rust-lang/regex/blob/master/regex-syntax/src/parser.rs#L1378
    """
    escaped_string = ""
    for c in s:
        if c in REGEX_PUNCT:
            escaped_string += "\\"
        escaped_string += c
    return escaped_string


# pylint:enable=line-too-long


def check_random_state(seed):
    """Turn seed into a :class:`numpy.random.RandomState` instance

    If seed is None, return the RandomState singleton used by np.random.
    If seed is an int, return a new RandomState instance seeded with seed.
    If seed is already a RandomState instance, return it.
    Otherwise raise ValueError.
    """
    # pylint: disable=W0212
    # pylint: disable=c-extension-no-member
    if seed is None or seed is np.random:
        return np.random.mtrand._rand  # pylint: disable=c-extension-no-member
    if isinstance(seed, (numbers.Integral, np.integer)):
        return np.random.RandomState(seed)
    if isinstance(seed, np.random.RandomState):
        return seed
    raise ValueError('%r cannot be used to seed a numpy.random.RandomState'
                     ' instance' % seed)


def get_slot_name_mapping(dataset, intent):
    """Returns a dict which maps slot names to entities for the provided intent
    """
    slot_name_mapping = dict()
    for utterance in dataset[INTENTS][intent][UTTERANCES]:
        for chunk in utterance[DATA]:
            if SLOT_NAME in chunk:
                slot_name_mapping[chunk[SLOT_NAME]] = chunk[ENTITY]
    return slot_name_mapping


def get_slot_name_mappings(dataset):
    """Returns a dict which maps intents to their slot name mapping"""
    return {intent: get_slot_name_mapping(dataset, intent)
            for intent in dataset[INTENTS]}


def ranges_overlap(lhs_range, rhs_range):
    if isinstance(lhs_range, dict) and isinstance(rhs_range, dict):
        return lhs_range[END] > rhs_range[START] \
               and lhs_range[START] < rhs_range[END]
    elif isinstance(lhs_range, (tuple, list)) \
            and isinstance(rhs_range, (tuple, list)):
        return lhs_range[1] > rhs_range[0] and lhs_range[0] < rhs_range[1]
    else:
        raise TypeError("Cannot check overlap on objects of type: %s and %s"
                        % (type(lhs_range), type(rhs_range)))



import io
import os

from setuptools import setup, find_packages

packages = [p for p in find_packages()
            if "tests" not in p and "debug" not in p]

PACKAGE_NAME = "snips_nlu"
ROOT_PATH = os.path.dirname(os.path.abspath(__file__))
README = os.path.join(ROOT_PATH, "README.rst")
PACKAGE_PATH = os.path.join(ROOT_PATH, PACKAGE_NAME)
VERSION = "__version__"

with io.open(os.path.join(PACKAGE_PATH, VERSION)) as f:
    version = f.readline().strip()

nlu_metrics_version = "0.12.0"

required = [
    "enum34==1.1.6",
    "numpy==1.14.0",
    "scipy==1.0.0",
    "scikit-learn==0.19.1",
    "sklearn-crfsuite==0.3.6",
    "semantic_version==2.6.0",
    "snips_nlu_utils==0.6.0",
    "snips_nlu_ontology==0.54.1",
    "num2words==0.5.6"
]

extras_require = {
    "doc": [
        "sphinx==1.7.1",
        "sphinxcontrib-napoleon==0.6.1",
        "sphinx-rtd-theme==0.2.4"
    ],
    "metrics": [
        "snips_nlu_metrics==%s" % nlu_metrics_version,
    ],
    "test": [
        "mock==2.0.0",
        "snips_nlu_metrics==%s" % nlu_metrics_version,
        "pylint==1.8.2",
        "coverage==4.4.2"
    ],
    "integration_test": [
        "snips_nlu_metrics==%s" % nlu_metrics_version,
        "snips_nlu_rust==0.55.0",
    ]
}

with io.open(README, encoding="utf8") as f:
    readme = f.read()

setup(name=PACKAGE_NAME,
      description="Snips Natural Language Understanding library",
      long_description=readme,
      version=version,
      author="Clement Doumouro, Adrien Ball",
      author_email="clement.doumouro@snips.ai, adrien.ball@snips.ai",
      license="Apache 2.0",
      install_requires=required,
      extras_require=extras_require,
      classifiers=[
          "Programming Language :: Python :: 2",
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 3.4",
          "Programming Language :: Python :: 3.5",
          "Programming Language :: Python :: 3.6",
      ],
      packages=packages,
      include_package_data=True,
      entry_points={
          "console_scripts": [
              "train-engine=cli.cli:main_train_engine",
              "engine-inference=cli.cli:main_engine_inference",
              "cross-val-metrics=cli.cli:main_cross_val_metrics",
              "train-test-metrics=cli.cli:main_train_test_metrics",
              "generate-dataset=snips_nlu_dataset:main_generate_dataset"
          ]
      },
      zip_safe=False)



from __future__ import unicode_literals

import os

# package
ROOT_PATH = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
RESOURCE_PACKAGE_NAME = "snips-nlu-resources"
PACKAGE_NAME = "snips_nlu"
RESOURCES_PATH = os.path.join(ROOT_PATH, PACKAGE_NAME, RESOURCE_PACKAGE_NAME)
PACKAGE_PATH = os.path.join(ROOT_PATH, PACKAGE_NAME)

# result
RES_INPUT = "input"
RES_INTENT = "intent"
RES_SLOTS = "slots"
RES_INTENT_NAME = "intentName"
RES_PROBABILITY = "probability"
RES_SLOT_NAME = "slotName"
RES_ENTITY = "entity"
RES_VALUE = "value"
RES_RAW_VALUE = "rawValue"
RES_MATCH_RANGE = "range"

# miscellaneous
AUTOMATICALLY_EXTENSIBLE = "automatically_extensible"
USE_SYNONYMS = "use_synonyms"
SYNONYMS = "synonyms"
DATA = "data"
INTENTS = "intents"
ENTITIES = "entities"
ENTITY = "entity"
ENTITY_KIND = "entity_kind"
SLOT_NAME = "slot_name"
TEXT = "text"
UTTERANCES = "utterances"
LANGUAGE = "language"
VALUE = "value"
BUILTIN_PARSER = "builtin_parser"
BUILTIN_PATH = "builtin_path"
BUILTIN_BINARY = "builtin_binary"
LABEL = "label"
RUSTLING_DIM_KIND = "rustling_dim_kind"
NGRAM = "ngram"
TOKEN_INDEXES = "token_indexes"
GAZETTEERS = "gazetteers"
STOP_WORDS = "stop_words"
STEMS = "stems"
WORD_CLUSTERS = "word_clusters"
SUPPORTED_LANGUAGES = "supported_languages"
CAPITALIZE = "capitalize"
VERBS_LEXEMES = "verbs_lexemes"
WORDS_INFLECTION = "words_inflection"
NOISE = "noise"
UNKNOWNWORD = "unknownword"
VALIDATED = "validated"
START = "start"
END = "end"

# builtin entities
SNIPS_AMOUNT_OF_MONEY = "snips/amountOfMoney"
SNIPS_DATETIME = "snips/datetime"
SNIPS_DURATION = "snips/duration"
SNIPS_NUMBER = "snips/number"
SNIPS_ORDINAL = "snips/ordinal"
SNIPS_PERCENTAGE = "snips/percentage"
SNIPS_TEMPERATURE = "snips/temperature"

# languages
LANGUAGE_DE = "de"
LANGUAGE_EN = "en"
LANGUAGE_ES = "es"
LANGUAGE_FR = "fr"
LANGUAGE_JA = "ja"
LANGUAGE_KO = "ko"



from snips_nlu.intent_classifier import LogRegIntentClassifier
from snips_nlu.intent_parser import DeterministicIntentParser
from snips_nlu.intent_parser import ProbabilisticIntentParser
from snips_nlu.nlu_engine.nlu_engine import SnipsNLUEngine
from snips_nlu.slot_filler import CRFSlotFiller

BUILTIN_NLU_PROCESSING_UNITS = [
    SnipsNLUEngine,
    ProbabilisticIntentParser,
    DeterministicIntentParser,
    LogRegIntentClassifier,
    CRFSlotFiller
]

NLU_PROCESSING_UNITS = {
    unit.unit_name: unit for unit in BUILTIN_NLU_PROCESSING_UNITS
}


def register_processing_unit(unit_type):
    """Allow to register a custom processing unit"""
    if unit_type.unit_name not in NLU_PROCESSING_UNITS:
        NLU_PROCESSING_UNITS[unit_type.unit_name] = unit_type


def reset_processing_units():
    """Remove all the custom processing units and reset to the default ones as
        defined in *BUILTIN_NLU_PROCESSING_UNITS*"""
    global NLU_PROCESSING_UNITS
    NLU_PROCESSING_UNITS = {
        unit.unit_name: unit for unit in BUILTIN_NLU_PROCESSING_UNITS
    }



# coding=utf-8
from __future__ import unicode_literals

import itertools
import re

from builtins import range
from builtins import str
from builtins import zip
from future.utils import iteritems
from snips_nlu_utils import normalize
from num2words import num2words

from snips_nlu.builtin_entities import get_builtin_entities
from snips_nlu.constants import (
    VALUE, RES_MATCH_RANGE, SNIPS_NUMBER, LANGUAGE_EN, LANGUAGE_ES,
    LANGUAGE_FR, LANGUAGE_DE, ENTITY, START, END)
from snips_nlu.languages import get_punctuation_regex, supports_num2words, \
    get_default_sep
from snips_nlu.tokenization import tokenize_light

AND_UTTERANCES = {
    LANGUAGE_EN: ["and", "&"],
    LANGUAGE_FR: ["et", "&"],
    LANGUAGE_ES: ["y", "&"],
    LANGUAGE_DE: ["und", "&"],
}

AND_REGEXES = {
    language: re.compile(
        r"|".join(r"(?<=\s)%s(?=\s)" % re.escape(u) for u in utterances),
        re.IGNORECASE)
    for language, utterances in iteritems(AND_UTTERANCES)
}


def build_variated_query(string, ranges_and_utterances):
    variated_string = ""
    current_ix = 0
    for rng, u in ranges_and_utterances:
        start = rng[START]
        end = rng[END]
        variated_string += string[current_ix:start]
        variated_string += u
        current_ix = end
    variated_string += string[current_ix:]
    return variated_string


def and_variations(string, language):
    and_regex = AND_REGEXES.get(language, None)
    variations = set()
    if and_regex is None:
        return variations

    matches = [m for m in and_regex.finditer(string)]
    if not matches:
        return variations

    matches = sorted(matches, key=lambda x: x.start())
    values = [({START: m.start(), END: m.end()}, AND_UTTERANCES[language])
              for m in matches]
    combinations = itertools.product(range(len(AND_UTTERANCES[language])),
                                     repeat=len(values))
    for c in combinations:
        ranges_and_utterances = [(values[i][0], values[i][1][ix])
                                 for i, ix in enumerate(c)]
        variations.add(build_variated_query(string, ranges_and_utterances))
    return variations


def punctuation_variations(string, language):
    variations = set()
    matches = [m for m in get_punctuation_regex(language).finditer(string)]
    if not matches:
        return variations

    matches = sorted(matches, key=lambda x: x.start())
    values = [({START: m.start(), END: m.end()}, (m.group(0), ""))
              for m in matches]

    combinations = itertools.product(range(2), repeat=len(matches))
    for c in combinations:
        ranges_and_utterances = [(values[i][0], values[i][1][ix])
                                 for i, ix in enumerate(c)]
        variations.add(build_variated_query(string, ranges_and_utterances))
    return variations


def digit_value(number_entity):
    value = number_entity[ENTITY][VALUE]
    if value == int(value):
        # Convert 24.0 into "24" instead of "24.0"
        value = int(value)
    return str(value)


def alphabetic_value(number_entity, language):
    value = number_entity[ENTITY][VALUE]
    if value != int(value):  # num2words does not handle floats correctly
        return None
    return num2words(value, lang=language)


def numbers_variations(string, language):
    variations = set()
    if not supports_num2words(language):
        return variations

    number_entities = get_builtin_entities(
        string, language, scope=[SNIPS_NUMBER])

    number_entities = sorted(number_entities,
                             key=lambda x: x[RES_MATCH_RANGE][START])
    if not number_entities:
        return variations

    digit_values = [digit_value(e) for e in number_entities]
    alpha_values = [alphabetic_value(e, language) for e in number_entities]

    values = [(n[RES_MATCH_RANGE], (d, a)) for (n, d, a) in
              zip(number_entities, digit_values, alpha_values)
              if a is not None]

    combinations = itertools.product(range(2), repeat=len(values))
    for c in combinations:
        ranges_and_utterances = [(values[i][0], values[i][1][ix])
                                 for i, ix in enumerate(c)]
        variations.add(build_variated_query(string, ranges_and_utterances))
    return variations


def case_variations(string):
    return {string.lower(), string.title()}


def normalization_variations(string):
    return {normalize(string)}


def flatten(results):
    return set(i for r in results for i in r)


def get_string_variations(string, language):
    variations = {string}
    variations.update(flatten(case_variations(v) for v in variations))
    variations.update(flatten(normalization_variations(v) for v in variations))
    # We re-generate case variations as normalization can produce new
    # variations
    variations.update(flatten(case_variations(v) for v in variations))
    variations.update(flatten(and_variations(v, language) for v in variations))
    variations.update(
        flatten(punctuation_variations(v, language) for v in variations))
    variations.update(
        flatten(numbers_variations(v, language) for v in variations))
    # Add single space variations
    single_space_variations = set(" ".join(v.split()) for v in variations)
    variations.update(single_space_variations)
    # Add tokenized variations
    tokenized_variations = set(
        get_default_sep(language).join(tokenize_light(v, language)) for v in
        variations)
    variations.update(tokenized_variations)
    return variations



from __future__ import print_function, absolute_import

import io
import itertools
import os

from builtins import object

from snips_nlu_dataset.builtin_entities import BuiltinEntity
from snips_nlu_dataset.custom_entities import CustomEntity, EntityUtterance
from snips_nlu.builtin_entities import is_builtin_entity


class IntentDataset(object):
    """Dataset of an intent

    Can parse utterances from a text file or an iterator.

    An example of utterance is:

        "the [role:role](president) of [country:country](France)"

    a Tag is in this format:

        [slot:entity_name](text_to_tag)

    Attributes:
        :class:IntentDataset.utterances: a list of :class:Utterance
        :class:IntentDataset.json: The dataset in json format
        :class:IntentDataset.intent_name: name of the intent
        :class:IntentDataset.language: language of the dataset
    """

    def __init__(self, intent_name, language):
        self.intent_name = intent_name
        self.language = language
        self.slots_by_name = dict()
        self.json_utterances = []
        self.utterances = []

    @classmethod
    def from_file(cls, language, file_name):
        intent_name = ".".join(os.path.basename(file_name).split('.')[:-1])
        with io.open(file_name) as f:
            lines = iter(l.strip() for l in f if l.strip() != '')
            return cls.from_iter(intent_name, language, lines)

    @classmethod
    def from_iter(cls, intent_name, language, samples_iter):
        """Generates a dataset from an iterator of samples

        :param intent_name: string
        :param language: string
        :param samples_iter: an iterator of string samples
        :return: :class:Dataset
        """
        dataset = cls(intent_name, language)
        for sample in samples_iter:
            u = Utterance.parse(sample)
            dataset.add(u)
        return dataset

    def add(self, utterance):
        """Adds an utterance to the dataset

        :param utterance: :class:Utterance
        """
        self.utterances.append(utterance)
        data = []
        for slot in utterance.slots:
            data.append(self.mk_slot(slot))
        self.json_utterances.append(dict(data=data))

    @property
    def json(self):
        """Dataset in json format"""
        return dict(language=self.language,
                    utterances=self.json_utterances,
                    entities=self.entities)

    @property
    def queries(self):
        """:return: An iter of all the example queries"""
        return (u.input for u in self.utterances)

    @property
    def entities(self):
        """retun all entities in json format for datasets"""
        ents = dict()
        for s in self.slots:
            if s.entity not in ents:
                ents[s.entity] = self.mk_entity(s)
            elif not is_builtin_entity(s.entity):
                ents[s.entity].utterances.append(EntityUtterance(s.text))
        return ents

    @classmethod
    def mk_entity(cls, slot, automatically_extensible=True, use_synonyms=True):
        if is_builtin_entity(slot.entity):
            return BuiltinEntity(slot.entity)
        return CustomEntity([EntityUtterance(slot.text)],
                            automatically_extensible, use_synonyms)

    @property
    def json_slots(self):
        """:return: slots but in json format"""
        return (self.intent_slot(s) for s in self.slots)

    @property
    def slots(self):
        """:return: an iter of the unique slots in the dataset"""
        all_slots = [
            s for u in self.utterances
            for s in u.slots
            if isinstance(s, Slot)
        ]
        all_slots.sort(key=lambda s: s.name)
        groups = itertools.groupby(all_slots, lambda s: s.entity)
        return (s for r, g in groups for s in g)

    @property
    def annotated(self):
        return (u.annotated for u in self.utterances)

    @staticmethod
    def intent_slot(slot):
        """json format expected in intent data"""
        return dict(
            name=slot.name,
            entity=slot.entity_name,
            entity_id=slot.entity
        )

    @staticmethod
    def mk_slot(slot):
        if isinstance(slot, Slot):
            return dict(
                slot_name=slot.name,
                entity=slot.entity,
                text=slot.text,
            )

        return dict(
            text=slot.text,
        )


class Utterance(object):
    def __init__(self, input, slots):
        self.input = input
        self.slots = slots

    @property
    def annotated(self):
        """Annotate with *

        :return: the sentence annotated just with stars
        >>> p = "the [role:snips/def](president) of [c:snips/def](France)"
        >>> u = Utterance.parse(p)
        >>> u.annotated
        u'the *president* of *France*'
        """
        binput = bytearray(self.input, 'utf-8')
        acc = 0
        star = ord('*')
        for slot in self.slots:
            if isinstance(slot, Slot):
                binput.insert(slot.range.start + acc, star)
                binput.insert(slot.range.end + acc + 1, star)
                acc += 2
        return binput.decode('utf-8')

    @staticmethod
    def stripped(input, slots):
        acc = 0
        s = ''
        new_slots = []
        for slot in slots:
            start = slot.range.start
            end = slot.range.end
            s += input[start:end]
            if isinstance(slot, Slot):
                acc += slot.tag_range.size
                range = Range(start - acc, end - acc)
                new_slot = Slot(slot.name, slot.entity, range, slot.text,
                                slot.tag_range)
                new_slots.append(new_slot)
                acc += 1
            else:
                range = Range(start - acc, end - acc)
                new_slots.append(Text(slot.text, range))
        return s, new_slots

    @staticmethod
    def parse(string):
        """Parses an utterance.

        :param string: an utterance in the Utterance format

        >>> u = Utterance.parse("president of [country:default](France)")
        >>> len(u.slots)
        2
        >>> u.slots[0].text
        'president of '
        >>> u.slots[0].range.start
        0
        >>> u.slots[0].range.end
        13
        """
        sm = SM(string)
        capture_text(sm)
        string, slots = Utterance.stripped(string, sm.slots)
        return Utterance(string, slots)


class Slot(object):
    def __init__(self, name, entity, range, text, tag_range):
        self.name = name
        self.entity = entity
        self.range = range
        self.text = text
        self.tag_range = tag_range


class Text(object):
    def __init__(self, text, range):
        self.text = text
        self.range = range


class Range(object):
    def __init__(self, start, end=None):
        self.start = start
        self.end = end

    @property
    def size(self):
        return self.end - self.start + 1


class SM(object):
    """State Machine for parsing"""

    def __init__(self, input):
        self.input = input
        self.slots = []
        self.current = 0

    def add_slot(self, slot_start, name, entity):
        """Adds a named slot

        :param slot_start: int position where the slot tag started
        :param name: string name of the slot
        :param entity: string entity id
        """
        tag_range = Range(slot_start - 1)
        slot = Slot(name=name, entity=entity, range=None, text=None,
                    tag_range=tag_range)
        self.slots.append(slot)

    def add_text(self, text):
        """Adds a simple text slot using the current position

        :param text: text of the slot
        """
        start = self.current
        end = start + len(text)
        slot = Text(text=text, range=Range(start=start, end=end))
        self.slots.append(slot)

    def add_tagged(self, text):
        """Adds this text to the last slot

        :param text: string the text that was tagged
        """
        assert self.slots
        slot = self.slots[-1]
        slot.text = text
        slot.tag_range.end = self.current - 1
        slot.range = Range(start=self.current,
                           end=self.current + len(text))

    def find(self, s):
        return self.input.find(s, self.current)

    def move(self, pos):
        """Moves the cursor of the state to position after given

        :param pos: int position to place the cursor just after
        """
        self.current = pos + 1

    def peek(self):
        return self[0]

    def read(self):
        c = self[0]
        self.current += 1
        return c

    def __getitem__(self, key):
        current = self.current
        if isinstance(key, int):
            return self.input[current + key]
        elif isinstance(key, slice):
            start = current + key.start if key.start else current
            return self.input[slice(start, key.stop, key.step)]
        else:
            raise Exception("Bad key")


def capture_text(state):
    next_pos = state.find('[')
    sub = state[:] if next_pos < 0 else state[:next_pos]

    if sub.strip():
        state.add_text(sub)
    if next_pos >= 0:
        state.move(next_pos)
        capture_slot(state)


def capture_slot(state):
    slot_start = state.current
    next_pos = state.find(':')
    if next_pos < 0:
        raise BadFormat()
    else:
        slot_name = state[:next_pos]
        state.move(next_pos)
        next_pos = state.find(']')
        if next_pos < 0:
            raise BadFormat()
        entity = state[:next_pos]
        state.move(next_pos)
        state.add_slot(slot_start, slot_name, entity)
        if state.read() != '(':
            raise BadFormat()
        capture_tagged(state)


def capture_tagged(state):
    next_pos = state.find(')')
    if next_pos < 1:
        raise BadFormat()
    else:
        tagged_text = state[:next_pos]
        state.add_tagged(tagged_text)
        state.move(next_pos)
        capture_text(state)


class BadFormat(Exception):
    pass



# coding=utf-8
from __future__ import unicode_literals, print_function

import argparse
import json
import os
from copy import deepcopy

from future.utils import iteritems

from snips_nlu_dataset.custom_entities import CustomEntity
from snips_nlu_dataset.intent_dataset import IntentDataset
from snips_nlu.builtin_entities import is_builtin_entity


class AssistantDataset(object):
    """Dataset of an assistant

    Merges a list of :class:AssistantDataset into a single dataset ready to be
    used by Snips NLU

    Attributes:
        :class:AssistantDataset.language: language of the dataset
        :class:AssistantDataset.intent_datasets: list of :class:IntentDataset
        :class:AssistantDataset.entities: dict of :class:CustomEntity
        :class:AssistantDataset.json: The dataset in json format
    """

    def __init__(self, language, intent_datasets, entities):
        self.language = language
        self.intent_datasets = intent_datasets
        self.entities = entities

    @classmethod
    def from_files(cls, language, intents_file_names=None,
                   entities_file_names=None):
        """Creates an :class:AssistantDataset from a language and a list of
        text files

        The assistant will associate each file to an intent, the name of the
        file being the intent name.
        """
        if intents_file_names is None:
            intents_file_names = []
        datasets = [IntentDataset.from_file(language, f) for f in
                    intents_file_names]
        if entities_file_names is None:
            entities_file_names = []
        entities = {
            os.path.splitext(os.path.basename(f))[0]: CustomEntity.from_file(f)
            for f in entities_file_names
        }
        return cls(language, datasets, entities)

    @property
    def json(self):
        intent_datasets_json = {d.intent_name: d.json
                                for d in self.intent_datasets}
        intents = {
            intent_name: {
                "utterances": dataset_json["utterances"]
            }
            for intent_name, dataset_json in iteritems(intent_datasets_json)
        }
        ents = deepcopy(self.entities)
        ents_values = dict()
        for entity_name, entity in iteritems(self.entities):
            ents_values[entity_name] = set(a.value for a in entity.utterances)
            if entity.use_synonyms:
                ents_values[entity_name].update(
                    set(t for s in entity.utterances for t in s.synonyms))

        for dataset in self.intent_datasets:
            for ent_name, ent in iteritems(dataset.entities):
                if ent_name not in ents:
                    ents[ent_name] = ent
                elif not is_builtin_entity(ent_name):
                    for u in ent.utterances:
                        if u.value not in ents_values:
                            ents[ent_name].utterances.append(u)
        ents = {
            entity_name: entity.json
            for entity_name, entity in iteritems(ents)
        }
        return dict(language=self.language, intents=intents, entities=ents)


def main_generate_dataset():
    parser = argparse.ArgumentParser(
        description="Create a Snips dataset from text friendly files",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--language", metavar="", type=str, default="en",
                        help="Language of the assistant")
    parser.add_argument("--intent-files", type=str, nargs="+",
                        help="List of intent files containing utterances")
    parser.add_argument("--entity-files", type=str, nargs="+",
                        help="List of entity files")
    args = parser.parse_args()
    dataset = AssistantDataset.from_files(args.language, args.intent_files,
                                          args.entity_files)
    print(json.dumps(dataset.json, indent=2))


if __name__ == '__main__':
    main_generate_dataset()



from builtins import map
from copy import deepcopy

from snips_nlu.pipeline.configs import ProcessingUnitConfig
from snips_nlu.pipeline.processing_unit import get_processing_unit_config
from snips_nlu.utils import classproperty


class NLUEngineConfig(ProcessingUnitConfig):
    """Configuration of a :class:`.SnipsNLUEngine` object

    Args:
        intent_parsers_configs (list): List of intent parser configs
            (:class:`.ProcessingUnitConfig`). The order in the list determines
            the order in which each parser will be called by the nlu engine.
    """

    # pylint: disable=super-init-not-called
    def __init__(self, intent_parsers_configs=None):

        if intent_parsers_configs is None:
            from snips_nlu.pipeline.configs import (
                ProbabilisticIntentParserConfig,
                DeterministicIntentParserConfig)
            intent_parsers_configs = [
                DeterministicIntentParserConfig(),
                ProbabilisticIntentParserConfig()
            ]
        self.intent_parsers_configs = list(map(get_processing_unit_config,
                                               intent_parsers_configs))

    # pylint: enable=super-init-not-called

    @classproperty
    def unit_name(cls):  # pylint:disable=no-self-argument
        from snips_nlu.nlu_engine.nlu_engine import SnipsNLUEngine
        return SnipsNLUEngine.unit_name

    def to_dict(self):
        return {
            "unit_name": self.unit_name,
            "intent_parsers_configs": [
                config.to_dict() for config in self.intent_parsers_configs
            ]
        }

    @classmethod
    def from_dict(cls, obj_dict):
        d = obj_dict
        if "unit_name" in obj_dict:
            d = deepcopy(obj_dict)
            d.pop("unit_name")
        return cls(**d)
